@@ -12,23 +12,24 @@ from enum import Enum
 from functools import lru_cache
 import numpy as np
 from .base import complex32
-from .config import FFT_CACHE_NB
+from . import config
 
 try:
     from .cuda import VkFFTApp as VkFFTApp_cuda, has_pycuda, has_cupy, vkfft_version
 
     if has_pycuda:
         import pycuda.gpuarray as cua
+        import pycuda.driver as cu_drv
     if has_cupy:
         import cupy as cp
-except ImportError:
+except (ImportError, OSError):
     has_cupy, has_pycuda = False, False
 
 try:
     from .opencl import VkFFTApp as VkFFTApp_cl, cla, vkfft_version
 
     has_opencl = True
-except ImportError:
+except (ImportError, OSError):
     has_opencl = False
 
 
@@ -40,7 +41,7 @@ class Backend(Enum):
     CUPY = 3
 
 
-def _prepare_transform(src, dest, cl_queue, r2c=False):
+def _prepare_transform(src, dest, cl_queue, cuda_stream, r2c=False):
     """
     Determine the backend from the input data.
     Create the destination array if necessary.
@@ -48,16 +49,26 @@ def _prepare_transform(src, dest, cl_queue, r2c=False):
     :param src: the source GPU array
     :param dest: the destination array. If None, a new GPU array is created.
     :param cl_queue: the opencl queue to use, or None
+    :param cuda_stream: the cuda stream to use (from cupy or pycuda), or None
     :param r2c: if True, this is for an R2C transform, so adapt the destination
         array accordingly.
-    :return: a tuple (backend, inplace, dest, cl_queue), also appending the
-    destination dtype for an r2c transform.
+    :return: a tuple (backend, inplace, dest, cl_queue, cuda_stream, devctx),
+        also appending the destination dtype for a r2c transform.
+        devctx is either the device or context unique ptr or id - this
+        is only used to cache the VkFFTapp (e.g. making sure the app is
+        re-instantiated if the cuda device changes). The cuda_stream
+        output will be the int ptr.
     """
     backend = Backend.UNKNOWN
+    fastidx = np.argmin(src.strides)  # fast axis is the last only for C-ordered arrays
+    if fastidx == src.ndim - 1:
+        order = 'C'
+    else:
+        order = 'F'
     if r2c:
         if src.dtype in [np.float16, np.float32, np.float64]:
             sh = list(src.shape)
-            sh[-1] = sh[-1] // 2 + 1
+            sh[fastidx] = sh[fastidx] // 2 + 1
             dtype = np.complex64
             if src.dtype == np.float16:
                 dtype = complex32
@@ -65,7 +76,7 @@ def _prepare_transform(src, dest, cl_queue, r2c=False):
                 dtype = np.complex128
         else:
             sh = list(src.shape)
-            sh[-1] = (sh[-1] - 1) * 2
+            sh[fastidx] = (sh[fastidx] - 1) * 2
             dtype = np.float32
             if src.dtype == complex32:
                 dtype = np.float16
@@ -73,7 +84,7 @@ def _prepare_transform(src, dest, cl_queue, r2c=False):
                 dtype = np.float64
     else:
         sh, dtype = None, None
-
+    devctx = None
     if has_pycuda:
         if isinstance(src, cua.GPUArray):
             backend = Backend.PYCUDA
@@ -82,10 +93,15 @@ def _prepare_transform(src, dest, cl_queue, r2c=False):
             src_ptr = int(src.gpudata)
             if dest is None:
                 if r2c:
-                    dest = cua.empty(tuple(sh), dtype=dtype, allocator=src.allocator)
+                    dest = cua.empty(tuple(sh), dtype=dtype, allocator=src.allocator, order=order)
                 else:
                     dest = cua.empty_like(src)
             dest_ptr = int(dest.gpudata)
+            if cuda_stream is None:
+                devctx = cu_drv.Context.get_current()
+            elif isinstance(cuda_stream, cu_drv.Stream):
+                # Pass an int to make sure it is hashable
+                cuda_stream = cuda_stream.handle
 
     if backend == Backend.UNKNOWN and has_opencl:
         if isinstance(src, cla.Array):
@@ -93,12 +109,13 @@ def _prepare_transform(src, dest, cl_queue, r2c=False):
             src_ptr = src.data.int_ptr
             if dest is None:
                 if r2c:
-                    dest = cla.empty(src.queue, tuple(sh), dtype=dtype, allocator=src.allocator)
+                    dest = cla.empty(src.queue, tuple(sh), dtype=dtype, allocator=src.allocator, order=order)
                 else:
                     dest = cla.empty_like(src)
             dest_ptr = dest.data.int_ptr
             if cl_queue is None:
                 cl_queue = src.queue
+            devctx = cl_queue.context
 
     if backend == Backend.UNKNOWN and has_cupy:
         if isinstance(src, cp.ndarray):
@@ -106,10 +123,15 @@ def _prepare_transform(src, dest, cl_queue, r2c=False):
             src_ptr = src.__cuda_array_interface__['data'][0]
             if dest is None:
                 if r2c:
-                    dest = cp.empty(tuple(sh), dtype=dtype)
+                    dest = cp.empty(tuple(sh), dtype=dtype, order=order)
                 else:
                     dest = cp.empty_like(src)
             dest_ptr = dest.__cuda_array_interface__['data'][0]
+            if cuda_stream is None:
+                cuda_stream = cp.cuda.get_current_stream()
+            if isinstance(cuda_stream, cp.cuda.Stream):
+                cuda_stream = cuda_stream.ptr
+            devctx = cp.cuda.Device().id
 
     if backend == Backend.UNKNOWN:
         raise RuntimeError("Could note determine the type of GPU array supplied, or the "
@@ -121,44 +143,61 @@ def _prepare_transform(src, dest, cl_queue, r2c=False):
     if r2c:
         if inplace:
             dest = src.view(dtype=dtype)
-        return backend, inplace, dest, cl_queue, dtype
+        return backend, inplace, dest, cl_queue, cuda_stream, devctx, dtype
     else:
-        return backend, inplace, dest, cl_queue
+        return backend, inplace, dest, cl_queue, cuda_stream, devctx
 
 
-@lru_cache(maxsize=FFT_CACHE_NB)
-def _get_fft_app(backend, shape, dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue):
+@lru_cache(maxsize=config.FFT_CACHE_NB)
+def _get_fft_app(backend, shape, dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue,
+                 devctx, strides=None, tune=False):
+    del devctx  # Variable is just used for proper lru_cache
+    sback = {Backend.PYCUDA: 'pycuda', Backend.CUPY: 'cupy', Backend.PYOPENCL: 'pyopencl'}[backend]
+    tune_config = {'backend': sback} if tune else None
     if backend in [Backend.PYCUDA, Backend.CUPY]:
         return VkFFTApp_cuda(shape, dtype, ndim=ndim, inplace=inplace,
-                             stream=cuda_stream, norm=norm, axes=axes)
+                             stream=cuda_stream, norm=norm, axes=axes, strides=strides,
+                             tune_config=tune_config)
     elif backend == Backend.PYOPENCL:
         return VkFFTApp_cl(shape, dtype, cl_queue, ndim=ndim, inplace=inplace,
-                           norm=norm, axes=axes)
+                           norm=norm, axes=axes, strides=strides,
+                           tune_config=tune_config)
 
 
-@lru_cache(maxsize=FFT_CACHE_NB)
-def _get_rfft_app(backend, shape, dtype, inplace, ndim, norm, cuda_stream, cl_queue):
+@lru_cache(maxsize=config.FFT_CACHE_NB)
+def _get_rfft_app(backend, shape, dtype, inplace, ndim, norm, cuda_stream, cl_queue,
+                  devctx, strides=None, tune=False):
+    del devctx  # Variable is just used for proper lru_cache
+    sback = {Backend.PYCUDA: 'pycuda', Backend.CUPY: 'cupy', Backend.PYOPENCL: 'pyopencl'}[backend]
+    tune_config = {'backend': sback} if tune else None
     if backend in [Backend.PYCUDA, Backend.CUPY]:
         return VkFFTApp_cuda(shape, dtype, ndim=ndim, inplace=inplace,
-                             stream=cuda_stream, norm=norm, r2c=True)
+                             stream=cuda_stream, norm=norm, r2c=True, strides=strides,
+                             tune_config=tune_config)
     elif backend == Backend.PYOPENCL:
         return VkFFTApp_cl(shape, dtype, cl_queue, ndim=ndim, inplace=inplace,
-                           norm=norm, r2c=True)
+                           norm=norm, r2c=True, strides=strides,
+                           tune_config=tune_config)
 
 
-@lru_cache(maxsize=FFT_CACHE_NB)
+@lru_cache(maxsize=config.FFT_CACHE_NB)
 def _get_dct_app(backend, shape, dtype, inplace, ndim, norm, dct_type,
-                 cuda_stream, cl_queue):
+                 cuda_stream, cl_queue, devctx, tune=False):
+    del devctx  # Variable is just used for proper lru_cache
+    sback = {Backend.PYCUDA: 'pycuda', Backend.CUPY: 'cupy', Backend.PYOPENCL: 'pyopencl'}[backend]
+    tune_config = {'backend': sback} if tune else None
     if backend in [Backend.PYCUDA, Backend.CUPY]:
         return VkFFTApp_cuda(shape, dtype, ndim=ndim, inplace=inplace,
-                             stream=cuda_stream, norm=norm, dct=dct_type)
+                             stream=cuda_stream, norm=norm, dct=dct_type,
+                             tune_config=tune_config)
     elif backend == Backend.PYOPENCL:
         return VkFFTApp_cl(shape, dtype, cl_queue, ndim=ndim, inplace=inplace,
-                           norm=norm, dct=dct_type)
+                           norm=norm, dct=dct_type,
+                           tune_config=tune_config)
 
 
 def fftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_queue=None,
-         return_scale=False):
+         return_scale=False, tune=False):
     """
     Perform a FFT on a GPU array, automatically creating the VkFFTApp
     and caching it for future re-use.
@@ -189,11 +228,19 @@ def fftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_queu
         the source array default queue will be used
     :param return_scale: if True, return the scale factor by which the result
         must be multiplied to keep its L2 norm after the transform
+    :param tune: if True, will activate the automatic tuning of VkFFT
+        parameters to maximise the FT throughput. This uses a quick
+        approach testing a few transforms (about 4) before choosing the
+        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
     :return: the destination array if return_scale is False, or (dest, scale)
     """
-    backend, inplace, dest, cl_queue = _prepare_transform(src, dest, cl_queue, False)
-    app = _get_fft_app(backend, src.shape, src.dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue)
-    app.fft(src, dest)
+    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
+    app = _get_fft_app(backend, src.shape, src.dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue, devctx,
+                       strides=src.strides, tune=tune)
+    if backend == Backend.PYOPENCL:
+        app.fft(src, dest, queue=cl_queue)
+    else:
+        app.fft(src, dest)
     if return_scale:
         s = app.get_fft_scale()
         return dest, s
@@ -201,7 +248,7 @@ def fftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_queu
 
 
 def ifftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_queue=None,
-          return_scale=False):
+          return_scale=False, tune=False):
     """
     Perform an inverse FFT on a GPU array, automatically creating the VkFFTApp
     and caching it for future re-use.
@@ -232,11 +279,19 @@ def ifftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_que
         the source array default queue will be used
     :param return_scale: if True, return the scale factor by which the result
         must be multiplied to keep its L2 norm after the transform
+    :param tune: if True, will activate the automatic tuning of VkFFT
+        parameters to maximise the FT throughput. This uses a quick
+        approach testing a few transforms (about 4) before choosing the
+        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
     :return: the destination array if return_scale is False, or (dest, scale)
     """
-    backend, inplace, dest, cl_queue = _prepare_transform(src, dest, cl_queue, False)
-    app = _get_fft_app(backend, src.shape, src.dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue)
-    app.ifft(src, dest)
+    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
+    app = _get_fft_app(backend, src.shape, src.dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue, devctx,
+                       strides=src.strides, tune=tune)
+    if backend == Backend.PYOPENCL:
+        app.ifft(src, dest, queue=cl_queue)
+    else:
+        app.ifft(src, dest)
     if return_scale:
         s = app.get_fft_scale()
         return dest, s
@@ -244,7 +299,7 @@ def ifftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_que
 
 
 def rfftn(src, dest=None, ndim=None, norm=1, cuda_stream=None, cl_queue=None,
-          return_scale=False):
+          return_scale=False, tune=False):
     """
     Perform a real->complex transform on a GPU array, automatically creating
     the VkFFTApp and caching it for future re-use.
@@ -277,13 +332,22 @@ def rfftn(src, dest=None, ndim=None, norm=1, cuda_stream=None, cl_queue=None,
         the source array default queue will be used
     :param return_scale: if True, return the scale factor by which the result
         must be multiplied to keep its L2 norm after the transform
+    :param tune: if True, will activate the automatic tuning of VkFFT
+        parameters to maximise the FT throughput. This uses a quick
+        approach testing a few transforms (about 4) before choosing the
+        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
     :return: the destination array if return_scale is False, or (dest, scale).
         For an in-place transform, the returned value is a view of the array
         with the appropriate type.
     """
-    backend, inplace, dest, cl_queue, dtype = _prepare_transform(src, dest, cl_queue, True)
-    app = _get_rfft_app(backend, src.shape, src.dtype, inplace, ndim, norm, cuda_stream, cl_queue)
-    app.fft(src, dest)
+    backend, inplace, dest, cl_queue, cuda_stream, devctx, dtype = \
+        _prepare_transform(src, dest, cl_queue, cuda_stream, True)
+    app = _get_rfft_app(backend, src.shape, src.dtype, inplace, ndim, norm, cuda_stream, cl_queue, devctx,
+                        strides=src.strides, tune=tune)
+    if backend == Backend.PYOPENCL:
+        app.fft(src, dest, queue=cl_queue)
+    else:
+        app.fft(src, dest)
     if return_scale:
         s = app.get_fft_scale()
         return dest.view(dtype=dtype), s
@@ -291,7 +355,7 @@ def rfftn(src, dest=None, ndim=None, norm=1, cuda_stream=None, cl_queue=None,
 
 
 def irfftn(src, dest=None, ndim=None, norm=1, cuda_stream=None, cl_queue=None,
-           return_scale=False):
+           return_scale=False, tune=False):
     """
     Perform a complex->real transform on a GPU array, automatically creating
     the VkFFTApp and caching it for future re-use.
@@ -324,20 +388,29 @@ def irfftn(src, dest=None, ndim=None, norm=1, cuda_stream=None, cl_queue=None,
         the source array default queue will be used
     :param return_scale: if True, return the scale factor by which the result
         must be multiplied to keep its L2 norm after the transform
+    :param tune: if True, will activate the automatic tuning of VkFFT
+        parameters to maximise the FT throughput. This uses a quick
+        approach testing a few transforms (about 4) before choosing the
+        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
     :return: the destination array if return_scale is False, or (dest, scale)
         For an in-place transform, the returned value is a view of the array
         with the appropriate type.
     """
-    backend, inplace, dest, cl_queue, dtype = _prepare_transform(src, dest, cl_queue, True)
-    app = _get_rfft_app(backend, dest.shape, dest.dtype, inplace, ndim, norm, cuda_stream, cl_queue)
-    app.ifft(src, dest)
+    backend, inplace, dest, cl_queue, cuda_stream, devctx, dtype = \
+        _prepare_transform(src, dest, cl_queue, cuda_stream, True)
+    app = _get_rfft_app(backend, dest.shape, dest.dtype, inplace, ndim, norm, cuda_stream, cl_queue, devctx,
+                        strides=dest.strides, tune=tune)
+    if backend == Backend.PYOPENCL:
+        app.ifft(src, dest, queue=cl_queue)
+    else:
+        app.ifft(src, dest)
     if return_scale:
         s = app.get_fft_scale()
         return dest.view(dtype=dtype), s
     return dest.view(dtype=dtype)
 
 
-def dctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_queue=None):
+def dctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_queue=None, tune=False):
     """
     Perform a real->real Direct Cosine Transform on a GPU array, automatically
     creating the VkFFTApp and caching it for future re-use.
@@ -361,16 +434,23 @@ def dctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_que
         for the transform. If None, the default one will be used
     :param cl_queue: the pyopencl.CommandQueue to be used. If None,
         the source array default queue will be used
+    :param tune: if True, will activate the automatic tuning of VkFFT
+        parameters to maximise the FT throughput. This uses a quick
+        approach testing a few transforms (about 4) before choosing the
+        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
     :return: the destination array.
     """
-    backend, inplace, dest, cl_queue = _prepare_transform(src, dest, cl_queue, False)
+    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
     app = _get_dct_app(backend, src.shape, src.dtype, inplace, ndim, norm,
-                       dct_type, cuda_stream, cl_queue)
-    app.fft(src, dest)
+                       dct_type, cuda_stream, cl_queue, devctx, tune=tune)
+    if backend == Backend.PYOPENCL:
+        app.fft(src, dest, queue=cl_queue)
+    else:
+        app.fft(src, dest)
     return dest
 
 
-def idctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_queue=None):
+def idctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_queue=None, tune=False):
     """
     Perform a real->real inverse Direct Cosine Transform on a GPU array,
     automatically creating the VkFFTApp and caching it for future re-use.
@@ -394,12 +474,19 @@ def idctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_qu
         for the transform. If None, the default one will be used
     :param cl_queue: the pyopencl.CommandQueue to be used. If None,
         the source array default queue will be used
+    :param tune: if True, will activate the automatic tuning of VkFFT
+        parameters to maximise the FT throughput. This uses a quick
+        approach testing a few transforms (about 4) before choosing the
+        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
     :return: the destination array.
     """
-    backend, inplace, dest, cl_queue = _prepare_transform(src, dest, cl_queue, False)
+    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
     app = _get_dct_app(backend, src.shape, src.dtype, inplace, ndim, norm,
-                       dct_type, cuda_stream, cl_queue)
-    app.ifft(src, dest)
+                       dct_type, cuda_stream, cl_queue, devctx, tune=tune)
+    if backend == Backend.PYOPENCL:
+        app.ifft(src, dest, queue=cl_queue)
+    else:
+        app.ifft(src, dest)
     return dest
 
 
