@@ -15,11 +15,22 @@ import unittest
 import time
 import timeit
 import socket
+import psutil
 import numpy as np
 from pyvkfft.test import TestFFT, TestFFTSystematic
 from pyvkfft.version import __version__, vkfft_version
 
 
+def cuda_info_str():
+    try:
+        # The functions used should not create a context on the GPU
+        from ..cuda import cuda_compile_version, cuda_runtime_version, cuda_driver_version
+        return "CUDA: driver %s, runtime %s, compiled %s" % \
+            (cuda_driver_version(), cuda_runtime_version(), cuda_compile_version())
+    except:
+        return ""
+
+
 def suite_default():
     suite = unittest.TestSuite()
     load_tests = unittest.defaultTestLoader.loadTestsFromTestCase
@@ -36,6 +47,16 @@ def suite_systematic():
 
 def make_html_pre_post(overwrite=False):
     if ('pyvkfft-test1000.html' not in os.listdir()) or overwrite:
+        # if this was called by a parent pyvkfft-test-suite, print that command
+        com = ""
+        for c in psutil.Process(os.getppid()).cmdline():
+            if "/" in c and "pyvkfft-test-suite" in c:
+                c = os.path.split(c)[-1]
+            com += "%s " % c
+        if 'pyvkfft-test-suite' in com:
+            com = 'Command:<tt>%s</tt>\n' % com
+        else:
+            com = ''
         # Need the html header, styles and the results' table beginning
         tmp = '<!DOCTYPE html>\n <html>\n <head> <style>\n' \
               'th, td { border: 1px solid grey;}\n' \
@@ -82,9 +103,9 @@ def make_html_pre_post(overwrite=False):
               '<body>\n' \
               '<div class="center">' \
               '<h2>pyVkFFT test results</h2>\n' \
-              '<h3>pyvkfft version: %s  VkFFFT version:%s host : %s</h3>\n' \
+              '<h3>pyvkfft: %s,  VkFFFT:%s %s host : %s</h3>\n' \
               '<div style="text-align:left;">' \
-              '<p>Methodology: the included graphs measure the accuracy of the forward ' \
+              '<p>%s<p>Methodology: the included graphs measure the accuracy of the forward ' \
               'and backward transforms: an array is generated with random uniform values ' \
               'between -0.5 and 0.5, and the results of its transform are compared ' \
               'with either pyfftw (in long double precision) if available, or scipy if ' \
@@ -93,7 +114,7 @@ def make_html_pre_post(overwrite=False):
               '<p>Note: for the R2C inverse transform, the result of the forward ' \
               'transform is used instead of re-using the random array (in order to have ' \
               'a proper half-Hermitian array), contrary to what is done for other ' \
-              'transforms. This explains with the IFFT R2C maximum (L<sub>&infin;</sub>) ' \
+              'transforms. This explains why the IFFT R2C maximum (L<sub>&infin;</sub>) ' \
               'errors are larger.' \
               '<p>Note 2: some "errors" for DCT may be due to unsupported sizes in VkFFT, ' \
               'which vary depending on the card and language used (amount of ' \
@@ -119,7 +140,8 @@ def make_html_pre_post(overwrite=False):
               '           <th>ERROR</th>' \
               '       </tr>\n' \
               '   </thead>\n' \
-              '<tbody class="center">\n' % (__version__, vkfft_version(), socket.gethostname())
+              '<tbody class="center">\n' % (__version__, vkfft_version(), cuda_info_str(),
+                                            socket.gethostname(), com)
         open("pyvkfft-test1000.html", "w").write(tmp)
     if ('pyvkfft-test1999.html' not in os.listdir()) or overwrite:
         tmp = '</tbody>\n' \
@@ -165,9 +187,7 @@ def name_next_file(pattern="pyvkfft-test%04d.html"):
     raise RuntimeError("name_next_file: '%s' files all used from 1001 to 1998. Maybe cleanup ?" % pattern)
 
 
-def main():
-    t0 = timeit.default_timer()
-    localt0 = time.localtime()
+def make_parser():
     epilog = "Examples:\n" \
              "   pyvkfft-test\n" \
              "      the regular test which tries the fft interface, using parallel\n" \
@@ -216,6 +236,11 @@ def main():
                              "which can be displayed by clicking on the type of transform.")
     parser.add_argument('--gpu', action='store',
                         help="Name (or sub-string) of the GPU to use")
+    parser.add_argument('--opencl_platform', action='store',
+                        help="Name (or sub-string) of the opencl platform to use (case-insensitive). "
+                             "Note that by default the PoCL platform is skipped, "
+                             "unless it is specifically requested or it is the only one available "
+                             "(PoCL has some issues with VkFFT for some transforms)")
     parser.add_argument('--mailto', action='store',
                         help="Email address the results will be sent to")
     parser.add_argument('--mailto_fail', action='store',
@@ -265,6 +290,9 @@ def main():
                         help="Use double precision (float64/complex128) instead of single")
     sysgrp.add_argument('--dry-run', action='store_true',
                         help="Perform a dry-run, printing the number of array shapes to test")
+    parser.add_argument('--fast-random', action='store', default=None,
+                        help="Use this option to run a random percentage of the full test suite, "
+                             "for faster results. A number between 5 and 100 is required.", type=int)
     sysgrp.add_argument('--inplace', action='store_true',
                         help="Use inplace transforms (NB: for R2C with ndim>=2, the x-axis "
                              "must be even-sized)")
@@ -301,6 +329,8 @@ def main():
                              "can be slower (or much slower on some architectures).")
     sysgrp.add_argument('--r2c', action='store_true', help="Test real-to-complex transform "
                                                            "(default is c2c)")
+    sysgrp.add_argument('--fstride', action='store_true',
+                        help="Test F-ordered arrays (default is C-ordered). Not supported for DCT")
     sysgrp.add_argument('--radix', action='store', nargs='*', type=int,
                         help="Perform only radix transforms. If no value is given, all available "
                              "radix transforms are allowed. Alternatively a list can be given: "
@@ -317,7 +347,7 @@ def main():
                         default=[2, 128])
     sysgrp.add_argument('--range-mb', action='store', nargs=2, type=int,
                         help="Allowed range of array sizes [min, max] in Mbytes, e.g. "
-                             "'--range 2 128'. This can be used to limit the arrays size "
+                             "'--range-mb 2 128'. This can be used to limit the arrays size "
                              "while allowing large lengths along individual dimensions. "
                              "It can also be used to separate runs with a given size range "
                              "and different nproc values. This takes into account the "
@@ -343,6 +373,13 @@ def main():
                         help="Change the timeout (in seconds) to raise a TimeOut error for "
                              "individual tests. After 4 have failed, give up.",
                         default=[120])
+    return parser
+
+
+def main():
+    t0 = timeit.default_timer()
+    localt0 = time.localtime()
+    parser = make_parser()
 
     # parser.print_help()
     args = parser.parse_args()
@@ -362,7 +399,11 @@ def main():
         t.db = args.db[0] if args.db is not None else None
         t.dry_run = args.dry_run
         t.dtype = np.float64 if args.double else np.float32
+        t.fast_random = args.fast_random
+        if args.fast_random is not None:
+            assert 5 <= args.fast_random <= 100, "--fast-random must be between 5 and 100"
         t.gpu = args.gpu
+        t.opencl_platform = args.opencl_platform
         t.graph = args.graph
         t.inplace = args.inplace
         t.lut = args.lut
@@ -371,7 +412,9 @@ def main():
         # t.ndims = args.ndims
         t.norm = args.norm[0]
         t.nproc = args.nproc[0]
+        t.opencl_platform = args.opencl_platform
         t.r2c = args.r2c
+        t.fstride = args.fstride
         t.radix = args.radix
         t.ref_long_double = args.ref_long_double
         t.max_pow = None if args.radix_max_pow is None else args.radix_max_pow[0]
@@ -404,6 +447,18 @@ def main():
         t.colour = args.colour
         t.gpu = args.gpu
         t.nproc = args.nproc[0]
+        t.opencl_platform = args.opencl_platform
+        t.vbackend = args.backend
+        if args.backend is not None:
+            # Remove tests depending on backend used
+            test_cuda_stream = False
+            for b in args.backend:
+                if 'pycuda' in b.lower():
+                    test_cuda_stream = True
+            if not test_cuda_stream:
+                del t.test_pycuda_streams
+            if 'pyopencl' not in args.backend:
+                del t.test_pyopencl_queues
         suite = unittest.defaultTestLoader.loadTestsFromTestCase(t)
         if t.verbose:
             res = unittest.TextTestRunner(verbosity=2).run(suite)
@@ -560,7 +615,6 @@ def main():
             html_out = name_next_file("pyvkfft-test%03d.html")
         with open(html_out, 'w') as f:
             f.write(html)
-
     sys.exit(int(nb_err_fail > 0))
 
 
